from .exceptions import PyAsiceError
from .ocsp import OCSP
from .tsa import TSA
from .xmlsig import XmlSignature


def finalize_signature(xml_signature: XmlSignature, issuer_cert: bytes, lt_ts=True, *, ocsp_url, tsa_url=None):
    """Finalize the XAdES signature in accordance with LT-TM profile, or LT-TS profile if `lt_ts` is True

    :param XmlSignature xml_signature:
    :param bytes issuer_cert:
    :param bool lt_ts: Whether to make the signature compliant with LT-TS and perform a TSA request
    :param ocsp_url:
    :param tsa_url: required if lt_ts is True
    """
    if lt_ts and not tsa_url:
        raise ValueError("TSA URL can not be empty when LT-TS profile is selected, requires a TSA service query")

    signature_value = xml_signature.get_signature_value()
    if not signature_value:
        raise PyAsiceError("The XML signature has not been signed properly")

    subject_cert = xml_signature.get_certificate()

    # Get an OCSP status confirmation
    ocsp = OCSP(ocsp_url)
    ocsp.validate(subject_cert, issuer_cert, signature_value)

    # Embed the OCSP response
    xml_signature.set_ocsp_response(ocsp)

    if lt_ts:
        # Get a signature TimeStamp
        tsa = TSA(tsa_url)
        tsr = tsa.get_timestamp(xml_signature.get_timestamped_message())
        xml_signature.set_timestamp_response(tsr)
    else:
        xml_signature.remove_timestamp_node()



import io
import os
import re
from typing import BinaryIO, Optional, Union
from zipfile import ZIP_DEFLATED, ZIP_STORED, ZipFile

from lxml import etree
from oscrypto.asymmetric import Certificate

from .exceptions import ContainerError, NoFilesToSign
from .xmlsig import XmlSignature


class Container(object):
    """
    Manipulate ASiC-E/BDoc v2 containers.

    Create a new container:

        bdoc = Container()
        bdoc\
            .add_file('test.pdf', b'Test data', 'application/pdf')\
            .add_signature(xmlsig)\
            .save('test.bdoc')

    Use `name` to open an existing container:

        bdoc = Container('test.bdoc')
        bdoc.verify_signatures()
        with bdoc.open_file('test.pdf') as f:
            assert f.read() == b'Test data'

        another_xmlsig = XmlSignature.create()...
        bdoc.add_signature(another_xmlsig).save()

    Spec: https://www.id.ee/public/bdoc-spec212-eng.pdf [1]
    """

    # save an `import ...` for users of the class
    Error = ContainerError
    NoFilesToSign = NoFilesToSign

    META_DIR = "META-INF"
    # > The names of these files shall contain the string "signatures" [1], ch.8
    SIGNATURE_FILES_REGEX = r"^%s/signatures(\d+)\.xml$" % META_DIR
    SIGNATURE_FILES_TEMPLATE = "%s/signatures{}.xml" % META_DIR

    # Manifest structure constants
    MANIFEST_FILE = "manifest.xml"
    MANIFEST_PATH = "{}/{}".format(META_DIR, MANIFEST_FILE)
    MANIFEST_NS = "urn:oasis:names:tc:opendocument:xmlns:manifest:1.0"
    MANIFEST_NAMESPACES = {
        "manifest": MANIFEST_NS,
    }
    MANIFEST_TAG_FILE_ENTRY = "{%s}file-entry" % MANIFEST_NS
    MANIFEST_ATTR_MEDIA_TYPE = "{%s}media-type" % MANIFEST_NS
    MANIFEST_ATTR_FULL_PATH = "{%s}full-path" % MANIFEST_NS

    MANIFEST_TEMPLATE_FILE = os.path.join(os.path.dirname(__file__), "templates", "manifest.xml")
    MIME_TYPE = "application/vnd.etsi.asic-e+zip"
    MIME_TYPE_FILE = "mimetype"

    def __init__(self, stream: Optional[BinaryIO] = None):
        """
        Create or open a BDOC/ASiC-E container from a stream, like a file handle or a BytesIO buffer.

        NOTE: if `stream` is not empty, opening an existing container is attempted.

        :param stream: a BytesIO buffer, or an open file
        """

        if stream is None:
            buffer = io.BytesIO()
            self._zip_file = ZipFile(buffer, "a")

            # add mimetype immediately, so that it's the first file in the TOC
            self._add_mimetype()

            # adding manifest can be deferred to the save() method call
            self._manifest = None
            self._manifest_write_required = True
        else:
            if isinstance(stream, io.BytesIO):
                buffer = stream
                buffer.seek(0)
            elif hasattr(stream, "read") and callable(stream.read):
                # Treat name_or_file as an open file handle
                buffer = io.BytesIO(stream.read())
            else:
                raise TypeError(f"Failed to open stream {type(stream)}")

            # create a zip file in 'append' mode, to make possible both reading and adding files
            self._zip_file = ZipFile(buffer, "a")

            self._verify_container_contents()
            self._manifest_write_required = False

        self._zip_buffer: io.BytesIO = buffer

    @classmethod
    def open(cls, path: str):
        with open(path, "rb") as f:
            return cls(f)

    def prepare_signature(self, signer_certificate: Union[bytes, Certificate]):
        """Generates an XML signature structure for files in the container"""
        if not self.has_data_files():
            raise NoFilesToSign(f"Container `{self}` contains no files to sign")

        # Generate a XAdES signature
        xml_sig = XmlSignature.create()

        for file_name, content, mime_type in self.iter_data_files():
            xml_sig.add_document(file_name, content, mime_type)

        xml_sig.set_certificate(signer_certificate).update_signed_info()

        return xml_sig

    def finalize(self) -> io.BytesIO:
        """Finalizes the zip archive and returns the BytesIO buffer.

        This allows passing the existing io.BytesIO handle without copying the data.

        **NOTE:** After this method call, the container becomes unusable. Otherwise,
          the buffer might be modified, which kind of defeats the optimization.
        """
        self._write_manifest()
        buffer = self._zip_buffer
        buffer.seek(0)
        self._close()
        return buffer

    def save(self, path: str):
        """Save the BDoc file to FS. This also closes the Container."""
        with open(path, "wb") as f:
            f.write(self.finalize().getbuffer())

    def add_file(self, file_name: str, binary_data: bytes, mime_type="application/octet-stream", compress=True):
        """Add a data file.

        :param file_name: the name of the file in the Zip archive
        :param binary_data: the content of the file
        :param mime_type: the file's content type as it appears in the container's manifest
        :param compress: on by default, there is no reason not to compress (except special cases)
        """
        if not self._zip_file:
            raise self.Error("Failed to add file: the container is closed.")
        manifest_xml = self._get_manifest_xml()
        new_manifest_entry = etree.Element(self.MANIFEST_TAG_FILE_ENTRY)
        new_manifest_entry.attrib[self.MANIFEST_ATTR_MEDIA_TYPE] = mime_type
        new_manifest_entry.attrib[self.MANIFEST_ATTR_FULL_PATH] = file_name
        manifest_xml.append(new_manifest_entry)
        compress_type = ZIP_DEFLATED if compress else ZIP_STORED
        self._zip_file.writestr(file_name, binary_data, compress_type)
        self._manifest_write_required = True
        return self

    @property
    def data_file_names(self):
        return [name for name, _ in self._enumerate_data_files()]

    @property
    def signature_file_names(self):
        return self._enumerate_signatures()

    def has_data_files(self):
        return any(self._enumerate_data_files())  # False if no elements

    def iter_data_files(self):
        """
        Iterate over 3-tuples of file name, content and mime_type
        """
        for file_name, mime_type in self._enumerate_data_files():
            with self.open_file(file_name) as f:
                yield file_name, f.read(), mime_type

    def open_file(self, file_name):
        """Read a file contained in the container"""
        if not self._zip_file:
            raise self.Error("Failed to add file: the container is closed.")
        return self._zip_file.open(file_name)

    def add_signature(self, signature: XmlSignature):
        """Add a signature calculated over the data files."""
        # Without this check, the zip container randomly gets corrupted
        # after adding a signature file, with errors like:
        # """ zipfile.BadZipFile: File name in directory 'META-INF/signatures1.xml'
        # """ and header b'META-INF/signatures2.xml' differ.
        self.verify_container()

        embedded_signatures = sorted(self._enumerate_signatures())

        if embedded_signatures:
            last_n = re.match(self.SIGNATURE_FILES_REGEX, embedded_signatures[-1]).group(1)
            next_n = int(last_n) + 1  # even with alphabetic file sorting, this gives valid next number
        else:
            next_n = 1

        new_sig_file = self.SIGNATURE_FILES_TEMPLATE.format(next_n)
        assert new_sig_file not in embedded_signatures
        self._zip_file.writestr(new_sig_file, signature.dump(), ZIP_DEFLATED)
        return self

    def iter_signatures(self):
        """Iterate over embedded signatures"""
        for entry in self._enumerate_signatures():
            with self.open_file(entry) as f:
                yield XmlSignature(f.read())

    def verify_signatures(self):
        """Verify all signatures in the container

        :raises signature_verifier.SignatureVerificationError:
        """
        for xmlsig in self.iter_signatures():
            xmlsig.verify()
        return self

    def verify_container(self):
        if not self._zip_file:
            raise self.Error("Failed to add file: the container is closed.")

        failed = self._zip_file.testzip()
        if failed:
            raise self.Error("The container contains errors. First broken file: %s" % failed)
        return self

    def _write_manifest(self):
        """Create/update the manifest"""
        if not self._manifest_write_required:
            return

        manifest_xml = self._get_manifest_xml()

        if self.MANIFEST_PATH in self._read_toc():
            self._delete_files(self.MANIFEST_PATH)

        self._zip_file.writestr(
            self.MANIFEST_PATH, b'<?xml version="1.0" encoding="UTF-8"?>' + etree.tostring(manifest_xml)
        )

    def _add_mimetype(self):
        # NOTE: the mimetype entry should be the first one and not compressed, as per ETSI TS 102 918 (though optional)
        self._zip_file.writestr(self.MIME_TYPE_FILE, self.MIME_TYPE.encode(), ZIP_STORED)

    def _read_toc(self):
        """Read table of contents"""
        if not self._zip_file:
            raise self.Error("Failed to add file: the container is closed.")
        return self._zip_file.namelist()

    def _get_manifest_xml(self):
        if self._manifest is None:
            # Create a manifest from template
            with open(self.MANIFEST_TEMPLATE_FILE, "rb") as f:
                self._manifest = etree.XML(f.read())
        return self._manifest

    def _enumerate_signatures(self):
        return [file_name for file_name in self._read_toc() if re.match(self.SIGNATURE_FILES_REGEX, file_name)]

    def _delete_files(self, *file_names_to_delete):
        new_buf = io.BytesIO()
        new_zip_file = ZipFile(new_buf, "a")
        file_names_to_delete = set(file_names_to_delete)
        for entry in self._zip_file.infolist():
            file_name = entry.filename
            if file_name in file_names_to_delete:
                file_names_to_delete.remove(file_name)
                continue

            with self.open_file(file_name) as f:
                new_zip_file.writestr(file_name, f.read(), entry.compress_type)

        self._zip_file.close()
        self._zip_buffer = new_buf
        self._zip_file = new_zip_file

    def _enumerate_data_files(self):
        """
        Yields 2-tuples of file name and mime_type
        """
        if not self._zip_file:
            raise self.Error("Failed to add file: the container is closed.")

        manifest_xml = self._get_manifest_xml()
        media_type_attr = self.MANIFEST_ATTR_MEDIA_TYPE
        full_path_attr = self.MANIFEST_ATTR_FULL_PATH

        for file_entry in manifest_xml.iterchildren():
            assert file_entry.tag == self.MANIFEST_TAG_FILE_ENTRY
            file_name = file_entry.attrib[full_path_attr]
            if file_name != "/":  # skip the 'root' entry
                yield file_name, file_entry.attrib[media_type_attr]

    def _verify_container_contents(self):
        self.verify_container()

        # Verify ZIP table of contents
        toc = self._read_toc()
        if not toc:
            raise self.Error(f"Empty container '{self}'")
        if toc[0] != self.MIME_TYPE_FILE:
            # Check that mimetype is the first entry.
            # NOTE: actually as per ETSI TS 102 918, MIME_TYPE_FILE is optional
            # neither is it explicitly stated as *the first entry* in BDOC2.1:2014 [OID: 1.3.6.1.4.1.10015.1000.3.2.3]
            # but digidoc software deems the opposite.
            raise self.Error(f"Container '{self}' must contain mime type file '{self.MIME_TYPE_FILE}' as first file")
        if self.MANIFEST_PATH not in toc:
            raise self.Error(f"Container '{self}' does not contain manifest file '{self.MANIFEST_FILE}'")

        # Read the meta data
        with self.open_file(self.MIME_TYPE_FILE) as f:
            mime_type = f.read()
        if mime_type.decode() != self.MIME_TYPE:
            raise self.Error(f"Invalid mime type '{mime_type}' for container '{self}'")

        try:
            with self.open_file(self.MANIFEST_PATH) as f:
                self._manifest = etree.XML(f.read())
        except Exception:
            raise self.Error(f"Failed to read manifest for container '{self}'")

        toc_data_files = [
            file_name
            for file_name in toc[1:]  # the first one is MIME_TYPE_FILE, can be skipped
            if not file_name.startswith(self.META_DIR)
        ]

        manifest_data_files = [name for name, _ in self._enumerate_data_files()]

        if sorted(toc_data_files) != sorted(manifest_data_files):
            raise self.Error("Manifest file is out of date")

    def _close(self):
        if self._zip_file:
            self._zip_file.close()
            self._zip_file = None
        self._zip_buffer = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._close()

    def __str__(self):
        return repr(self)



class PyAsiceError(Exception):
    """
    A generic exception that can happen while dealing with ASic-E/BDoc 2 files/signatures
    """

    pass


class ContainerError(PyAsiceError):
    pass


class NoFilesToSign(PyAsiceError):
    pass


class SignatureVerificationError(PyAsiceError):
    pass


class InvalidSignatureAlgorithm(SignatureVerificationError):
    pass



import base64
import copy
import hashlib
import logging
import os
from datetime import datetime
from typing import Optional, Union

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from lxml import etree
from oscrypto.asymmetric import Certificate, load_certificate

from .exceptions import SignatureVerificationError
from .ocsp import OCSP
from .signature_verifier import verify
from .tsa import TSA


logger = logging.getLogger(__name__)


def get_utc_time():
    # For testing purposes, as we can't patch a datetime object
    return datetime.utcnow()


class XmlSignature:
    """
    Usage:

        # Create a signature XAdES structure
        sig = XmlSignature.create() \
            .add_document('test.pdf', b'PDF data', 'application/pdf') \
            .set_certificate(file_or_binary_data) \
            .update_signed_info()

        # Get the actual signature from e.g. smartid, out of scope
        signature_value = sign(id_code, sig.digest())

        # Get OCSP and TSA confirmation
        result_xml = sig.set_signature_value(signature_value) \
            .verify() \
            .set_ocsp_response(...) \
            .set_timestamp_response(...) \
            .dump()

        # or:
        from .utils import finalize_signature
        finalize_signature(sig, lt_ts=True)
    """

    SIGNATURE_TEMPLATE = os.path.join(os.path.dirname(__file__), "templates", "signature.xml")
    NAMESPACES = {
        "asic": "http://uri.etsi.org/02918/v1.2.1#",
        "ds": "http://www.w3.org/2000/09/xmldsig#",
        "xades": "http://uri.etsi.org/01903/v1.3.2#",
    }

    C14N_METHODS = (
        "http://www.w3.org/TR/2001/REC-xml-c14n-20010315",  # this is the REQUIRED c14n algorithm [xmldsig]
        "http://www.w3.org/2001/10/xml-exc-c14n#",
        "http://www.w3.org/2006/12/xml-c14n11",
    )

    DIGEST_ALGORITHMS = {
        "sha256": "http://www.w3.org/2001/04/xmlenc#sha256",
    }

    # https://www.w3.org/TR/xmldsig-core1/#sec-AlgID
    SIGNATURE_ALGO_ID_TEMPLATE = "http://www.w3.org/2001/04/xmldsig-more#{algo}"
    SIGNATURE_ALGORITHMS = [
        "rsa-sha256",  # the default one, is embedded in the XML template.
        "ecdsa-sha256",
        "rsa-sha384",
        "ecdsa-sha384",
        "rsa-sha512",
        "ecdsa-sha512",
    ]

    SIGNED_PROPERTIES_TYPE = (
        # Standards are ambiguous about this:
        "http://uri.etsi.org/01903#SignedProperties",  # BDOC 2.1.2 mandates this
        "http://uri.etsi.org/01903/v1.1.1#SignedProperties",  # and this is as per https://www.w3.org/TR/XAdES/
    )

    NEW_SIGNATURE_ID = "S1"  # This is arbitrary but used a few times in the XAdES structure.

    def __init__(self, xml_or_binary_data):
        if isinstance(xml_or_binary_data, (etree._Element, etree._ElementTree)):
            self.xml = xml_or_binary_data
        else:
            parser = etree.XMLParser(remove_blank_text=True, remove_comments=True)
            try:
                self.xml = etree.XML(xml_or_binary_data, parser=parser)
            except ValueError:
                logger.exception("Failed to load XML document: %s", xml_or_binary_data)
                raise

        if self.xml.tag != "{%s}XAdESSignatures" % self.NAMESPACES["asic"]:
            raise ValueError("Expecting an 'asic:XAdESSignatures' root node")

        data_objects_props_node = self._get_node("ds:SignedInfo")
        doc_entries = data_objects_props_node.findall('ds:Reference[@Type=""]', namespaces=self.NAMESPACES)
        self.doc_ids = [doc_entry.attrib["Id"] for doc_entry in doc_entries]

        self._certificate: Certificate = None
        self._prepared = None

    @classmethod
    def create(cls):
        """Create a XAdES structure from the accompanying template"""
        with open(cls.SIGNATURE_TEMPLATE, "rb") as f:
            xml_sig = cls(f.read().replace(b"{SIGNATURE_ID}", cls.NEW_SIGNATURE_ID.encode("ascii")))
            xml_sig.doc_ids = []
            return xml_sig

    def dump(self):
        return b'<?xml version="1.0" encoding="UTF-8"?>' + etree.tostring(self.xml)

    def get_signing_time(self):
        time_node = self._get_signing_time_node()
        if time_node is None or not time_node.text:
            return None
        return time_node.text

    def get_certificate(self) -> Certificate:
        if not self._certificate:
            cert_asn1 = self.get_certificate_value()
            if cert_asn1:
                # cache it on the instance, may be used a few times
                self._certificate = load_certificate(cert_asn1)
        return self._certificate

    def get_certificate_value(self):
        cert_node = self._get_node("ds:X509Certificate")
        if cert_node is None or not cert_node.text:
            return None
        return base64.b64decode(cert_node.text)

    def get_certificate_issuer_common_name(self):
        subject_cert = self.get_certificate()
        return subject_cert.asn1.issuer.native["common_name"] if subject_cert else None

    def set_certificate(self, subject_cert: Union[bytes, Certificate]):
        """Set the signer's certificate

        :param subject_cert: bytes, file name (Python 3.4+), asn1crypto.x509.Certificate objects
        """
        if not isinstance(subject_cert, Certificate):
            subject_cert = load_certificate(subject_cert)

        # cache it on the instance, may be used a few times
        self._certificate = subject_cert

        cert_asn1 = subject_cert.asn1
        der_encoded_cert = cert_asn1.dump()
        serial_number = ("%d" % cert_asn1.serial_number).encode("ascii")

        cert_node = self._get_node("ds:X509Certificate")
        cert_node.text = base64.b64encode(der_encoded_cert)

        cert_props = self._get_node("xades:SigningCertificate")
        cert_props.find(".//ds:DigestValue", self.NAMESPACES).text = base64.b64encode(cert_asn1.sha256)
        cert_props.find(".//ds:X509SerialNumber", self.NAMESPACES).text = serial_number

        # No idea what value is possible, but rfc4514 is most common, so get it from a cryptography object
        x509_cert = x509.load_der_x509_certificate(der_encoded_cert, default_backend())
        cert_props.find(".//ds:X509IssuerName", self.NAMESPACES).text = x509_cert.issuer.rfc4514_string()

        return self

    def add_document(self, file_name, binary_data, mime_type, hash_type="sha256") -> "XmlSignature":
        """Add a document for signing

        :param file_name: the file name to display in the container
        :param mime_type: the document's mime type
        :param binary_data: the document's contents
        :param hash_type: the hash function to use for digesting
        :return:
        """
        try:
            digest_algo = self.DIGEST_ALGORITHMS[hash_type]
        except KeyError as e:
            raise ValueError("Unknown hash type: %s" % hash_type) from e

        digest_fn = getattr(hashlib, hash_type)
        doc_hash = digest_fn(binary_data).digest()

        signed_info = self._get_node("ds:SignedInfo")
        first_ref_entry = signed_info.find('.//ds:Reference[@Type=""]', self.NAMESPACES)
        doc_id = first_ref_entry.attrib["Id"]

        doc_props = self._get_node("xades:SignedDataObjectProperties")
        first_doc_entry = doc_props.find('.//xades:DataObjectFormat[@ObjectReference="#%s"]' % doc_id, self.NAMESPACES)

        if self.doc_ids:
            next_num = len(self.doc_ids) + 1

            # generate new Id attribute
            while True:
                new_doc_id = "r-id-{}".format(next_num)
                if new_doc_id not in self.doc_ids:
                    break
                next_num += 1

            # Instead of manually creating elements, just copy the structure
            new_ref_entry = copy.deepcopy(first_ref_entry)
            signed_info.append(new_ref_entry)

            new_doc_entry = copy.deepcopy(first_doc_entry)
            doc_props.append(new_doc_entry)
        else:
            new_doc_id = doc_id.format(DOCUMENT_NUMBER=1)
            new_doc_entry = first_doc_entry
            new_ref_entry = first_ref_entry

        self.doc_ids.append(new_doc_id)
        new_ref_entry.attrib["Id"] = new_doc_id
        new_ref_entry.attrib["URI"] = file_name
        new_ref_entry.find(".//ds:DigestMethod", self.NAMESPACES).attrib["Algorithm"] = digest_algo
        new_ref_entry.find(".//ds:DigestValue", self.NAMESPACES).text = base64.b64encode(doc_hash)

        new_doc_entry.attrib["ObjectReference"] = "#%s" % new_doc_id
        new_doc_entry.find(".//xades:MimeType", self.NAMESPACES).text = mime_type

        return self

    def update_signed_info(self) -> "XmlSignature":
        """Calculate the digest over SignedProperties and embed it in SignedInfo"""

        # Set signing time
        time_node = self._get_signing_time_node()
        # Add a UTC timestamp. Can't use isoformat() as it adds +00:00 and microseconds
        #  which can break the parser elsewhere
        time_node.text = get_utc_time().strftime("%Y-%m-%dT%H:%M:%SZ")

        self._calc_signed_properties_hash(update=True)

        self._prepared = True
        return self

    @property
    def prepared(self) -> bool:
        if self._prepared is None:
            self._prepared = False
            if self.get_signing_time():
                old, new = self._calc_signed_properties_hash()
                if old == new:
                    self._prepared = True
        return self._prepared

    def signed_data(self) -> bytes:
        if not self.prepared:
            raise ValueError("The XML signature is not prepared")
        sign_info_node = self._get_node("ds:SignedInfo")
        return self.canonicalize(sign_info_node)

    def digest(self) -> bytes:
        signature_algo = self.get_signature_algorithm()
        hash_algo_name = signature_algo.split("-")[-1]
        hash_algo = getattr(hashlib, hash_algo_name)
        return hash_algo(self.signed_data()).digest()

    def verify(self) -> "XmlSignature":
        hash_algo = self.get_signature_algorithm().split("-")[-1]
        cert = self.get_certificate_value()
        signature = self.get_signature_value()
        signed_data = self.signed_data()
        verify(cert, signature, signed_data, hash_algo)
        return self

    def get_signature_value(self):
        sig_value_node = self._get_signature_value_node()
        try:
            text = sig_value_node.text.strip()
        except AttributeError:
            return None

        return base64.b64decode(text) if text else None

    def set_signature_value(self, signature: bytes) -> "XmlSignature":
        """Insert the base64-encoded value of a signature obtained from a signing service or device

        NOTE: the signature method should be known in advance, as it's part of the SignedInfo structure over which
          the signature is calculated.

        :param signature: Binary signature
        """
        sig_value_node = self._get_signature_value_node()
        sig_value_node.text = base64.b64encode(signature)
        return self

    def get_signature_algorithm(self) -> str:
        """
        Returns the #hash part of the corresponding node, in form of 'rsa-sha256'

        This algorithm is expected to be present in SIGNATURE_ALGORITHMS and be in form of
        {CRYPTOMETHOD}-{DIGESTMETHOD}. See the note for `set_signature_algorithm` below.
        """
        sig_method_node = self._get_node("ds:SignatureMethod")
        return sig_method_node.attrib["Algorithm"].split("#")[-1]

    def set_signature_algorithm(self, algo: str = None) -> "XmlSignature":
        """Set a signature algorithm, if it is not the default one (rsa-sha256).

        NOTE: Since the algorithm is included in the signed data, it is not possible to change the algo
        after signing.

        Ultimately, if an external signing service MAY select different algos at its discretion,
        this field should be ignored.

        FWIW the verification method of this class only takes into account
        the hash algo (DIGESTMETHOD) part of this attribute.

        :param algo: signature algorithm, one of SIGNATURE_ALGORITHMS
        """
        if algo:
            algo = algo.lower()
            if algo not in self.SIGNATURE_ALGORITHMS:
                raise ValueError("Unsupported signature algorithm")
        else:
            algo = self.SIGNATURE_ALGORITHMS[0]
        sig_method_node = self._get_node("ds:SignatureMethod")
        sig_method_node.attrib["Algorithm"] = self.SIGNATURE_ALGO_ID_TEMPLATE.format(algo=algo)
        return self

    def set_root_ca_cert(self, root_cert: Union[Certificate, bytes]) -> "XmlSignature":
        """Sets a root CA cert. This is not mandatory

        :param root_cert: can be a PEM- or DER-encoded bytes content, or an `oscrypto.Certificate` object
        """
        certs_node = self._get_node("xades:CertificateValues")
        ca_node = etree.Element("{%s}EncapsulatedX509Certificate" % self.NAMESPACES["xades"])
        ca_node.attrib["Id"] = "%s-ROOT-CA-CERT" % self.NEW_SIGNATURE_ID
        if not isinstance(root_cert, Certificate):
            root_cert = load_certificate(root_cert)
        ca_node.text = base64.b64encode(root_cert.asn1.dump())
        certs_node.append(ca_node)
        return self

    def set_ocsp_response(self, ocsp_response: OCSP, embed_ocsp_certificate=False) -> "XmlSignature":
        """
        Embed the OCSP response and certificates

        :param OCSP ocsp_response:
        :param bool embed_ocsp_certificate: Whether to add ocsp certificate to the xml.
            This is needed when the OCSP service in use
            does not embed the certificate in its response.
        :return: self
        """
        ocsp_response_node = self._get_node("xades:EncapsulatedOCSPValue")
        ocsp_response_node.text = base64.b64encode(ocsp_response.get_encapsulated_response())

        if embed_ocsp_certificate:
            ocsp_certs_node = self._get_node("xades:CertificateValues")
            ocsp_certs = ocsp_response.get_responder_certs()
            cert_node = ocsp_certs_node.find(".//xades:EncapsulatedX509Certificate", namespaces=self.NAMESPACES)
            cert_node.text = base64.b64encode(ocsp_certs[0].dump())
            cert_node.attrib["Id"] = "S1-Responder-cert-1"

            for i, next_cert in enumerate(ocsp_certs[1:]):
                cert_node = copy.deepcopy(cert_node)
                cert_node.text = base64.b64encode(ocsp_certs[next_cert].dump())
                cert_node.attrib["Id"] = "S1-Responder-cert-%d" % i
                ocsp_certs_node.append(cert_node)

        return self

    def get_ocsp_response(self) -> Optional[OCSP]:
        ocsp_response_node = self._get_node("xades:EncapsulatedOCSPValue")
        try:
            text = ocsp_response_node.text.strip()
        except AttributeError:
            return None

        return OCSP.load(base64.b64decode(text)) if text else None

    def verify_ocsp_response(self) -> "XmlSignature":
        """Verify embedded OCSP response.

        :raises exceptions.SignatureVerificationError:
        """
        try:
            self.get_ocsp_response().verify()
        except AttributeError:
            raise SignatureVerificationError("The XML Signature doesn't contain an OCSP response")
        return self

    def verify_ts_response(self) -> "XmlSignature":
        """Verify embedded TSA response.

        :raises exceptions.SignatureVerificationError:
        """
        TSA.verify(self.get_timestamp_response(), self.get_timestamped_message())
        return self

    def get_timestamped_message(self) -> bytes:
        sig_value_node = self._get_signature_value_node()
        method = self.get_c14n_method("xades:SignatureTimeStamp")
        return self.canonicalize(sig_value_node, method)

    def get_timestamp_response(self) -> Optional[bytes]:
        """
        Get the embedded TSA response for an LT-TS profile signature.

        LT-TM must not even have the XML node.
        """
        ts_value_node = self._get_node("xades:EncapsulatedTimeStamp")
        try:
            text = ts_value_node.text.strip()
        except AttributeError:
            return None
        return base64.b64decode(text) if text else None

    def set_timestamp_response(self, tsr) -> "XmlSignature":
        ts_value_node = self._get_node("xades:EncapsulatedTimeStamp")
        ts_value_node.text = base64.b64encode(tsr.dump())
        return self

    def remove_timestamp_node(self) -> "XmlSignature":
        """LT-TM profile requires the absence of the XML node"""
        ts_value_node = self._get_node("xades:SignatureTimeStamp")
        ts_value_node.getparent().remove(ts_value_node)
        return self

    def get_c14n_method(self, parent_node="ds:SignedInfo"):
        """Get a c14n method used within a specific context given by `parent_node`

        The default context is the SignedInfo node. Also encountered in SignatureTimestamp
        """
        method_node = self._get_node("{}/ds:CanonicalizationMethod".format(parent_node))
        if method_node is not None:
            method = method_node.attrib["Algorithm"]
            if method not in self.C14N_METHODS:
                raise ValueError("Unknown c14n method: {}".format(method))
        else:
            method = self.C14N_METHODS[0]
        return method

    def canonicalize(self, node, method=None) -> bytes:
        if method is not None:
            assert method in self.C14N_METHODS
        else:
            method = self.get_c14n_method()
        exclusive = "xml-exc-c14n" in method
        return etree.tostring(node, method="c14n", exclusive=exclusive)

    def _get_node(self, tag_name) -> etree._Element:
        return self.xml.find(".//{}".format(tag_name), namespaces=self.NAMESPACES)

    def _get_signed_properties_node(self):
        return self.xml.find(
            "./ds:Signature/ds:Object/xades:QualifyingProperties/xades:SignedProperties", self.NAMESPACES
        )

    def _get_signing_time_node(self):
        return self.xml.find(
            "./ds:Signature/ds:Object/xades:QualifyingProperties/xades:SignedProperties/"
            "xades:SignedSignatureProperties/xades:SigningTime",
            self.NAMESPACES,
        )

    def _get_signed_info_ref_node(self):
        """Find the SignedInfo/Reference node that refers to the XMLSig itself (not to the signed files)"""
        return next(
            self.xml.find('./ds:Signature/ds:SignedInfo/ds:Reference[@Type="%s"]' % ref_type, self.NAMESPACES)
            for ref_type in self.SIGNED_PROPERTIES_TYPE
        )

    def _get_signed_properties_c14n_algo(self, signed_info_ref_node: etree._Element):
        """
        Gets a transform/c14n algorithm for SignedProperties

        The child nodes may be absent, in this case use the default c14n algo

        # This is very obscure in the standard:
        # https://www.w3.org/TR/2002/REC-xmldsig-core-20020212/#sec-ReferenceProcessingModel
        """
        try:
            c14n_algo = signed_info_ref_node.find("./ds:Transforms/ds:Transform", self.NAMESPACES).attrib["Algorithm"]
        except:  # noqa: E722
            c14n_algo = None
        return c14n_algo

    def _get_signature_value_node(self):
        return self.xml.find("./ds:Signature/ds:SignatureValue", self.NAMESPACES)

    def _calc_signed_properties_hash(self, update=False) -> tuple:
        """
        Calculates, and updates, if requested, the SignedInfo/Reference/DigestValue

        ...based on the current SignedProperties

        :param update: update the digest value with current SignedProperties hash
        :return: tuple(DigestValue node value, SignedProperties hash value)
        """
        si_ref_node = self._get_signed_info_ref_node()

        signed_props_node = self._get_signed_properties_node()
        c14n_algo = self._get_signed_properties_c14n_algo(si_ref_node)

        signed_props_c14n = self.canonicalize(signed_props_node, c14n_algo)
        # TODO select algorithm based on DigestMethod // update DigestMethod
        signed_props_hash = hashlib.sha256(signed_props_c14n).digest()

        new_digest_value = base64.b64encode(signed_props_hash).decode()

        si_digest_node = si_ref_node.find("./ds:DigestValue", self.NAMESPACES)
        if update:
            si_digest_node.text = new_digest_value

        return si_digest_node.text, new_digest_value



import hashlib

import requests
from asn1crypto.cms import ContentInfo  # noqa
from asn1crypto.core import OctetString, SetOf
from asn1crypto.tsp import TimeStampReq, TimeStampResp

from .exceptions import PyAsiceError
from .signature_verifier import verify


class TSAError(PyAsiceError):
    pass


class TSA:
    """
    Query a Time Stamping Authority (TSA) for a signature time stamp
    """

    REQUEST_CONTENT_TYPE = "application/timestamp-query"
    RESPONSE_CONTENT_TYPE = "application/timestamp-reply"

    def __init__(self, url):
        self.url = url
        self.ts_response = None

    def get_timestamp(self, message: bytes) -> ContentInfo:
        """Get the time stamp structure for embedding in a XAdES signature

        How to prepare the message:
        https://www.etsi.org/deliver/etsi_ts/101900_101999/101903/01.04.02_60/ts_101903v010402p.pdf
        section 7.3
        """
        request = self.build_ts_request(message)

        try:
            response = requests.post(
                self.url,
                data=request.dump(),
                headers={
                    "Content-Type": self.REQUEST_CONTENT_TYPE,
                    "Connection": "close",
                },
            )
            response.raise_for_status()
        except requests.ConnectionError as e:
            raise TSAError("Failed to connect to TSA service at {}".format(self.url)) from e
        except requests.HTTPError as e:
            raise TSAError("Bad response from TSA service at {}: {}".format(self.url, e)) from e

        content_type = response.headers["Content-Type"]
        if content_type != self.RESPONSE_CONTENT_TYPE:
            raise TSAError(f"Invalid response content type '{content_type}' returned by OCSP service at {self.url}")

        ts_response = TimeStampResp.load(response.content)
        # see asn1crypto.tsp.PKIStatus
        status = ts_response["status"]["status"].native
        if status != "granted":
            raise TSAError(f"Timestamping service denied the request with a status of {status}")

        self.ts_response = ts_response
        return ts_response["time_stamp_token"]

    @classmethod
    def verify(cls, ts_response: bytes, original_message: bytes):
        """
        Verify that the signature in the response is valid.

        https://tools.ietf.org/html/rfc5652#section-5.4
        """
        try:
            # any error during data structure parsing means the TS response is not valid
            ts_response = ContentInfo.load(ts_response)
            ts_content_type = ts_response.native["content_type"]

            # see asn1crypto.cms.ContentType
            if ts_content_type != "signed_data":
                raise ValueError(f"The TS response has invalid content type '{ts_content_type}'")

            content = ts_response["content"]

            # Get the encapsulated TS message ("econtent")
            econtent = content["encap_content_info"]["content"]

            # verify that the response refers to the original message

            message_imprint = econtent.native["message_imprint"]
            original_imprint = cls.build_message_imprint(original_message)
            if (
                message_imprint["hash_algorithm"]["algorithm"] != original_imprint["hash_algorithm"]["algorithm"]
                or message_imprint["hashed_message"] != original_imprint["hashed_message"]
            ):
                raise ValueError("The timestamped message differs from the original one")

            # Verify that "econtent" hash matches the message_digest attribute of signed_attrs

            digest_algo = content["digest_algorithms"][0]["algorithm"].native
            econtent_digest = getattr(hashlib, digest_algo)(econtent.contents).digest()

            signer_info = content["signer_infos"][0]

            signed_attrs = signer_info["signed_attrs"]

            md = [a for a in signed_attrs.native if a["type"] == "message_digest"][0]
            the_hash = md["values"][0]
            if econtent_digest != the_hash:
                raise ValueError("Message digests do not match")

            # Verify the signature with the included cert

            # To get signed data, we need some magic, as quoted from the RFC 5652 (the link in docstring):
            #    the DER encoding of the EXPLICIT SET OF tag, rather than of the IMPLICIT [0]
            #    tag, MUST be included in the message digest calculation
            # i.e. replace the current (class, tag) pair of the struct with those of the SET-OF type
            signed_attrs.implicit = False
            signed_attrs.class_ = SetOf.class_
            signed_attrs.tag = SetOf.tag
            signed_data = signed_attrs.dump(True)

            cert = content["certificates"][0].chosen

            signature: OctetString = signer_info["signature"]
            signature_bytes = signature.native

            sig_algo = signer_info["signature_algorithm"]["algorithm"].native
            sig_hash_algo = sig_algo.split("_")[0]

            verify(cert.dump(), signature_bytes, signed_data, sig_hash_algo)

        except Exception as e:
            raise TSAError("Invalid TSA response format") from e
        return ts_response

    @staticmethod
    def build_message_imprint(message):
        if not isinstance(message, bytes):
            raise TypeError(f"Invalid message type: expected bytes, got {type(message)}")
        if not message:
            raise ValueError("Message can not be empty")

        return {
            "hash_algorithm": {"algorithm": "sha256"},
            "hashed_message": hashlib.sha256(message).digest(),
        }

    @classmethod
    def build_ts_request(cls, message):
        return TimeStampReq(
            {
                "version": "v1",
                "message_imprint": cls.build_message_imprint(message),
                "cert_req": True,  # Need the TSA cert in the response for validation
            }
        )



#!/usr/bin/env python

from setuptools import setup


with open("pyasice/__init__.py") as f:
    version_line = next(line for line in f if line.startswith("__version__"))

version = version_line.split("=")[-1].strip().strip('"' + "'")

readme = open("README.md").read()

requirements = [line for line in open("requirements.txt").readlines() if line and not line.startswith("#")]

setup(
    name="pyasice",
    version=version,
    description="""Manipulate ASiC-E containers and XAdES/eIDAS signatures for Estonian e-identity services""",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Thorgate",
    author_email="yuriy@thorgate.eu",
    url="https://github.com/thorgate/pyasice",
    packages=[
        "pyasice",
    ],
    include_package_data=True,
    install_requires=requirements,
    license="ISC",
    keywords="esteid asice xades smartid smart-id mobiilid mobile-id idcard",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)
